/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.smlvalidators;

import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.cosmos.rm.validation.internal.artifacts.ElementNode;
import org.eclipse.cosmos.rm.validation.internal.common.IValidationOutput;
import org.eclipse.cosmos.rm.validation.internal.core.AbstractSMLValidator;
import org.eclipse.cosmos.rm.validation.internal.databuilders.ComplexTypeElementBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.DataBuilderRegistry;
import org.eclipse.cosmos.rm.validation.internal.databuilders.ElementTypeMapDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.RefTypeElementBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.RootElementAliasLookupBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.SubstitutionBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TargetElementBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TargetElementInstanceBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TargetRequiredBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TargetRequiredInstanceBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TargetTypeBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TargetTypeInstanceBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TypeInheritanceDataBuilderImpl;
import org.eclipse.cosmos.rm.validation.internal.databuilders.XSElementBuilder;
import org.eclipse.cosmos.rm.validation.internal.util.ParserHelper;

/**
 * This class is used to validate the sml:targetElement, sml:targetRequired and sml:targetType sml extensions. 
 * @author sleeloy
 *
 */
public class TargetValidator extends AbstractSMLValidator
{

	public void initialize(Map validationAttribute)
	{
		super.initialize(validationAttribute);
		DataBuilderRegistry builderRegistry = DataBuilderRegistry.instance();
		
		builderRegistry.registerDataStructureBuilder(TypeInheritanceDataBuilderImpl.ID, new TypeInheritanceDataBuilderImpl());
		builderRegistry.registerDataStructureBuilder(TargetElementBuilder.ID, new TargetElementBuilder());
		builderRegistry.registerDataStructureBuilder(TargetTypeBuilder.ID, new TargetTypeBuilder());
		builderRegistry.registerDataStructureBuilder(RefTypeElementBuilder.ID, new RefTypeElementBuilder());
		builderRegistry.registerDataStructureBuilder(ComplexTypeElementBuilder.ID, new ComplexTypeElementBuilder());
		builderRegistry.registerDataStructureBuilder(TargetElementInstanceBuilder.ID, new TargetElementInstanceBuilder());
		builderRegistry.registerDataStructureBuilder(TargetTypeInstanceBuilder.ID, new TargetTypeInstanceBuilder());
		builderRegistry.registerDataStructureBuilder(RootElementAliasLookupBuilder.ID, new RootElementAliasLookupBuilder());		
		builderRegistry.registerDataStructureBuilder(XSElementBuilder.ID, new XSElementBuilder());
		builderRegistry.registerDataStructureBuilder(SubstitutionBuilder.ID, new SubstitutionBuilder());
		builderRegistry.registerDataStructureBuilder(ElementTypeMapDataBuilder.ID, new ElementTypeMapDataBuilder());
		builderRegistry.registerDataStructureBuilder(TargetRequiredBuilder.ID, new TargetRequiredBuilder());
		builderRegistry.registerDataStructureBuilder(TargetRequiredInstanceBuilder.ID, new TargetRequiredInstanceBuilder());
		
		
		
						
	}

	public boolean validate()
	{
		IValidationOutput validationLogger = getValidationOutput();
		IDataBuilder targetElement = DataBuilderRegistry.instance().getDataStructureBuilder(TargetElementBuilder.ID);
		IDataBuilder targetType = DataBuilderRegistry.instance().getDataStructureBuilder(TargetTypeBuilder.ID);
		IDataBuilder refTypeElement = DataBuilderRegistry.instance().getDataStructureBuilder(RefTypeElementBuilder.ID);
		IDataBuilder typeInheritance = DataBuilderRegistry.instance().getDataStructureBuilder(TypeInheritanceDataBuilderImpl.ID);
		IDataBuilder complexTypeElement = DataBuilderRegistry.instance().getDataStructureBuilder(ComplexTypeElementBuilder.ID);
		IDataBuilder targetInstanceElement  = DataBuilderRegistry.instance().getDataStructureBuilder(TargetElementInstanceBuilder.ID);
		IDataBuilder targetInstanceType  = DataBuilderRegistry.instance().getDataStructureBuilder(TargetTypeInstanceBuilder.ID);
		IDataBuilder rootElementAliasElement  = DataBuilderRegistry.instance().getDataStructureBuilder(RootElementAliasLookupBuilder.ID);
		IDataBuilder xsElementElement  = DataBuilderRegistry.instance().getDataStructureBuilder(XSElementBuilder.ID);
		IDataBuilder substitionElement  = DataBuilderRegistry.instance().getDataStructureBuilder(SubstitutionBuilder.ID);
		IDataBuilder elementTypeMapBuilder  = DataBuilderRegistry.instance().getDataStructureBuilder(ElementTypeMapDataBuilder.ID);
		IDataBuilder targetRequiredBuilder  = DataBuilderRegistry.instance().getDataStructureBuilder(TargetRequiredBuilder.ID);			
		IDataBuilder targetRequiredInstanceBuilder  = DataBuilderRegistry.instance().getDataStructureBuilder(TargetRequiredInstanceBuilder.ID);			
								
		Map targetElems = (Map)targetElement.getDataStructure();
		Map targetTypes = (Map)targetType.getDataStructure();
		Map refTypeElems = (Map)refTypeElement.getDataStructure();
		Map complexTypeElems = (Map)complexTypeElement.getDataStructure();
		Map inheritance = (Map)typeInheritance.getDataStructure();
		Map targetInstanceElems = (Map)targetInstanceElement.getDataStructure();
		Map targetInstanceTypes = (Map)targetInstanceType.getDataStructure();
		Map rootElementAliasElems = (Map)rootElementAliasElement.getDataStructure();
		Map xsElementElems = (Map)xsElementElement.getDataStructure();
		Map substitionElementElems = (Map)substitionElement.getDataStructure();
		Map elementTypeMapBuilderStructure = (Map)elementTypeMapBuilder.getDataStructure();
		Map targetRequiredBuilderStructure = (Map)targetRequiredBuilder.getDataStructure();
		Map targetRequiredInstanceBuilderStructure = (Map)targetRequiredInstanceBuilder.getDataStructure();
		
		//The sml:targetElement attribute is supported on element 
		//declarations whose type is sml:refType or a derived type
		//of sml:refType. 
		Iterator iter = targetElems.keySet().iterator();
		while (iter.hasNext()){
			Object targetElem = iter.next();
			//check if elem is of type sml:refType
			if (refTypeElems.get(targetElem) == null){
				boolean valid = false;
				//check inheritance tree
				Object parent = inheritance.get(targetElem);
				while (parent != null){
					if (refTypeElems.get(parent) != null){
						valid = true;
						break;
					}
					parent = inheritance.get(parent);
					
				}
				if (!valid){
					validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetElementNotRefType,targetElem));
					//stop validating
					return false;
				}
			}
			
			//make sure the target value is a global type
			String targetValue = (String)targetElems.get(targetElem);
			targetValue = ParserHelper.removeNameSpace(targetValue);
			
			//get type
			ElementNode xsElement = (ElementNode)xsElementElems.get(targetValue);
			String type = ParserHelper.removeNameSpace(xsElement.getType());
			//check if the type is a global type
			if (complexTypeElems.get(type) == null){
				
				boolean valid = false;
				//check inheritance tree
				String parent = (String)inheritance.get(type);
				parent = ParserHelper.removeNameSpace(parent);
				while (parent != null){
					if (complexTypeElems.get(parent) != null){
						valid = true;
						break;
					}
					parent = (String)inheritance.get(parent);
					parent = ParserHelper.removeNameSpace(parent);
					
				}
				if (!valid){
					validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetElementNotDefined,targetElem,parent));
					//stop validating
					return false;
				}
				if (!valid){
					validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetElementNotDefined, targetElem, targetValue));
					//stop validating
					return false;
				}
			}
		}
		
		
		Iterator iter1 = targetInstanceElems.keySet().iterator();
		while (iter1.hasNext()){
			ElementNode key = (ElementNode)iter1.next();
			//get GTE
			Object GTE = ParserHelper.removeNameSpace((String)targetElems.get(key.getName()));
			//get target ref.  Note:This should not be null
			Object targetRef = targetInstanceElems.get(key);
			//check to see if the target ref is instance of GTE
			if (targetRef.equals(GTE)){
				continue;
			}
			else{
				//the ref could be an alias
				//Search through alias datastructure for actual element
				Iterator iter2 = rootElementAliasElems.keySet().iterator();
				while (iter2.hasNext()){
					List aliasList = (List)iter2.next();
					if (aliasList.contains(targetRef)){
						targetRef = rootElementAliasElems.get(aliasList);
						break;
					}
				}
				if (targetRef.equals(GTE)){
					continue;
				}
				//there could be a substition group
				Object substitionGroup = substitionElementElems.get(targetRef);				
				boolean valid = false;
				while (substitionGroup != null){
					substitionGroup = ParserHelper.removeNameSpace((String)substitionGroup);
					if (substitionGroup.equals(GTE)){
						valid = true;
						break;
					}			
					substitionGroup = substitionElementElems.get(substitionGroup);
				}
				if (valid) continue;
			}			
			validationLogger.reportError(ValidationMessages.bind(ValidationMessages.invalidTargetInstance, targetRef, GTE));		 //$NON-NLS-2$
			return false;
		}
		
		
		/*****Target Type*****/
		//The sml:targetType attribute is supported on element 
		//declarations whose type is sml:refType or a derived type
		//of sml:refType. 
		iter = targetTypes.keySet().iterator();
		while (iter.hasNext()){
			Object targetTypeElem = iter.next();
			//check if elem is of type sml:refType
			if (refTypeElems.get(targetTypeElem) == null){
				boolean valid = false;
				//check inheritance tree
				Object parent = inheritance.get(targetTypeElem);
				while (parent != null){
					if (refTypeElems.get(parent) != null){
						valid = true;
						break;
					}
					parent = inheritance.get(parent);
					
				}
				if (!valid){
					validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetElementNotRefType,targetTypeElem));
					//stop validating
					return false;
				}
			}
			
			//make sure the target value is a global type
			String targetValue = (String)targetTypes.get(targetTypeElem);
			
			//get type
			String type = ParserHelper.removeNameSpace(targetValue);
			//check if the type is a global type
			if (complexTypeElems.get(type) == null){
				
				boolean valid = false;
				//check inheritance tree
				String parent = (String)inheritance.get(type);
				parent = ParserHelper.removeNameSpace(parent);
				while (parent != null){
					if (complexTypeElems.get(parent) != null){
						valid = true;
						break;
					}
					parent = (String)inheritance.get(parent);
					parent = ParserHelper.removeNameSpace(parent);
					
				}
				if (!valid){
					validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetElementNotDefined,targetTypeElem,parent));
					//stop validating
					return false;
				}
				if (!valid){
					validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetElementNotDefined, targetTypeElem, targetValue));
					//stop validating
					return false;
				}
			}
		}		
			
		
		//Make sure that the target type
		iter1 = targetInstanceTypes.keySet().iterator();
		while (iter1.hasNext()){
			ElementNode key = (ElementNode)iter1.next();
			//get GTE
			Object GTEType = ParserHelper.removeNameSpace((String)targetTypes.get(key.getName()));
			//get target ref.  Note:This should not be null
			Object targetRef = targetInstanceTypes.get(key);
			Object originalRef = targetRef;
			//check to see if the target ref is instance of GTE
			Object targetRefTypeNS = elementTypeMapBuilderStructure.get(targetRef);
			String targetRefType = ParserHelper.removeNameSpace((String)targetRefTypeNS);
			if (GTEType.equals(targetRefType)){
				continue;
			}
			else{
				//check inheritance tree
				if (targetRefTypeNS != null){
					String parent = (String)inheritance.get(targetRefTypeNS);
					parent = ParserHelper.removeNameSpace(parent);
					boolean valid = false;
					while (parent != null){
						if (GTEType.equals(parent)){
							valid = true;
							break;
						}
						parent = (String)inheritance.get(parent);
						parent = ParserHelper.removeNameSpace(parent);
						
					}					
					if (valid) 
						continue;
				}
				//the ref could be an alias
				//Search through alias datastructure for actual element
				Iterator iter2 = rootElementAliasElems.keySet().iterator();
				while (iter2.hasNext()){
					List aliasList = (List)iter2.next();
					if (aliasList.contains(targetRef)){
						targetRef = rootElementAliasElems.get(aliasList);
						break;
					}
				}
				targetRefTypeNS = elementTypeMapBuilderStructure.get(targetRef);
				targetRefType = ParserHelper.removeNameSpace((String)targetRefTypeNS);
				if (GTEType.equals(targetRefType)){
					continue;
				}
				//check inheritance tree
				if (targetRefTypeNS != null){
					String parent = (String)inheritance.get(targetRefTypeNS);
					parent = ParserHelper.removeNameSpace(parent);
					boolean valid = false;
					while (parent != null){
						if (GTEType.equals(parent)){
							valid = true;
							break;
						}
						parent = (String)inheritance.get(parent);
						parent = ParserHelper.removeNameSpace(parent);
						
					}					
					if (valid) 
						continue;
				}				
				//there could be a substition group
				Object substitionGroup = substitionElementElems.get(targetRef);				
				boolean valid = false;
				while (substitionGroup != null){
					substitionGroup = ParserHelper.removeNameSpace((String)substitionGroup);
					
					targetRefType =ParserHelper.removeNameSpace((String)elementTypeMapBuilderStructure.get(substitionGroup));
					if (GTEType.equals(targetRefType)){
						valid = true;
						break;
					}			
					substitionGroup = substitionElementElems.get(substitionGroup);
				}
				if (valid) continue;
			}			
			Object[] errorParams = new String[5];
			errorParams[0] = key.getName();
			errorParams[1] = key.getLine()+""; //$NON-NLS-1$
			errorParams[2] = key.getColumn()+""; //$NON-NLS-1$
			errorParams[3] = originalRef;
			errorParams[4] = GTEType;
			validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetTypeInvalid,errorParams));
			return false;
		}				
		
		//targetRequired
		//The sml:targetRequired attribute is supported on element 
		//declarations whose type is sml:refType or a derived type
		//of sml:refType. 
		iter = targetRequiredBuilderStructure.keySet().iterator();
		while (iter.hasNext()){
			Object targetRequired = iter.next();
			//check if elem is of type sml:refType
			if (refTypeElems.get(targetRequired) == null){
				boolean valid = false;
				//check inheritance tree
				Object parent = inheritance.get(targetRequired);
				while (parent != null){
					if (refTypeElems.get(parent) != null){
						valid = true;
						break;
					}
					parent = inheritance.get(parent);
					
				}
				if (!valid){
					validationLogger.reportError(ValidationMessages.bind(ValidationMessages.targetRequiredNotRefType,targetRequired));
					//stop validating
					return false;
				}
			}
			
		}
		
		iter1 = targetRequiredInstanceBuilderStructure.keySet().iterator();
		while (iter1.hasNext()){
			ElementNode key = (ElementNode)iter1.next();
			//get target ref.  Note:This should not be null
			Object targetRequiredURI = targetRequiredInstanceBuilderStructure.get(key);
			boolean uriValid = isURIValid((String)targetRequiredURI);
			
			//if uri is valid then go to the next element since it doesn't matter if the 
			//targetRequired attribute is true or false
			if (uriValid)
				continue;

			//get target required attribute value.  Note:This should not be null
			Object elementType = elementTypeMapBuilderStructure.get(key.getName());
			Object requiredValue = targetRequiredBuilderStructure.get(elementType);
			
			if (requiredValue != null){
				if ("true".equals(requiredValue)){
					validationLogger.reportError(key.getName() + " can not define a URI that is dangling or null.  "+elementType+" defines attribute sml:targetRequired to true.");		 //$NON-NLS-2$
					return false;					
				}
				else{
					//uri can be dangling or null
					continue;
				}
			}
			else{
				//check inheritance tree
				if (elementType != null){
					String parent = (String)inheritance.get(elementType);
					parent = ParserHelper.removeNameSpace(parent);
					boolean valid = false;
					while (parent != null){
						requiredValue = targetRequiredBuilderStructure.get(parent);
						if (requiredValue != null){
							if ("true".equals(requiredValue)){
								validationLogger.reportError(key.getName() + " can not define a URI that is dangling or null.  "+elementType+" defines attribute sml:targetRequired to true.");		 //$NON-NLS-2$
								return false;					
							}
							else{
								//uri can be dangling or null
								valid = true;
								break;
							}
						}						
						parent = (String)inheritance.get(parent);
						parent = ParserHelper.removeNameSpace(parent);
						
					}					
					if (valid) 
						continue;
				}
				
				//there could be a substition group
				Object substitionGroup = substitionElementElems.get(key.getName());				
				boolean valid = false;
				while (substitionGroup != null){
					substitionGroup = ParserHelper.removeNameSpace((String)substitionGroup);
					
					elementType = ParserHelper.removeNameSpace((String)elementTypeMapBuilderStructure.get(substitionGroup));
					requiredValue = targetRequiredBuilderStructure.get(elementType);
					if ("true".equals(requiredValue)){
						validationLogger.reportError(key.getName() + " can not define a URI that is dangling or null.  "+elementType+" defines attribute sml:targetRequired to true.");		 //$NON-NLS-2$
						return false;					
					}
					else if (requiredValue != null){
						valid = true;
						//uri can be dangling or null
						break;
					}
					substitionGroup = substitionElementElems.get(substitionGroup);
				}
				if (valid) continue;
			}			
			validationLogger.reportError(key.getName() + " can not define a URI that is dangling or null.  "+elementType+" defines attribute sml:targetRequired to true.");		 //$NON-NLS-2$
			return false;
		}
		
		
							
		return true;
	}

	protected boolean isURIValid(String uri){
		return false;
	}
}
